"""Usage: brilcalc trg [options] 

options:
  -h,  --help                  Show this screen.
  -c CONNECT                   DB Service name [default: offline]
  -p AUTHPATH                  Authentication file
  -r RUN                       Run number
  -o OUTPUTFILE                Output csv file. Special file '-' for stdout.
  --pathinfo                   Show hltpath, l1seed expression
  --prescale                   Show trigger prescale changes
  --hltpath HLTPATH            hltpath name/pattern 
  --hltconfig HLTCONFIG        HLT configid or hltkey/pattern  
  --ignore-mask                Ignore trigger bit masks
  --output-style OSTYLE        Screen output style. tab, html, csv [default: tab]
 
"""

import os
from docopt import docopt
from schema import Schema
from brilws.cli import clicommonargs

def validate(optdict):
    result={}
    argdict = clicommonargs.argvalidators
    #extract sub argdict here
    myvalidables = ['-c','-r','--hltpath','--hltconfig',str]
    argdict = dict((k,v) for k,v in clicommonargs.argvalidators.iteritems() if k in myvalidables)
    schema = Schema(argdict)
    result = schema.validate(optdict)
    return result

if __name__ == '__main__':
    print(docopt(__doc__))



from datetime import datetime, timedelta,date
import time

nbx = 3564
orbits_per_ls = 262144
bunchspace_us = 0.02495
bunchspace_s = 24.95e-09

def OrbitDuration(norbits = 1):
    return timedelta(microseconds=(nbx*norbits*bunchspace_us))

def LSDuration(norbitsperls = orbits_per_ls):
    return OrbitDuration(norbitsperls)

def OrbitToTime(begtimetuple,orbitnumber,begorbit=0):
    return begtimetuple+OrbitDuration(orbitnumber-begorbit)

def DatetimeToSec(dt, epoch=datetime(1970,1,1)):
    td = dt - epoch
    return int(td.total_seconds())

def SecToDatetime(tssecs,utc=True):
    if utc:
        return datetime.utcfromtimestamp(int(tssecs))
    else:
        return datetime.fromtimestamp(int(tssecs))
    
def SecToDatetimeStr(tssecs,utc=True,fmt='%m/%d/%y %H:%M:%S'):
    return SecToDatetime(tssecs,utc=utc).strftime(fmt)

def WeekOfYear(dt):
    return date(dt.year,dt.month,dt.day).isocalendar()[1]
    
if __name__ == '__main__':
    tsec = DatetimeToSec(datetime.utcnow())
    print tsec
    t = SecToDatetime(tsec)
    print t
    print WeekOfYear(t)
    tp = t.timetuple()
    print tp.tm_year,tp.tm_mon,tp.tm_wday,tp.tm_yday
    print SecToDatetimeStr(tsec)
    print LSDuration()
    print LSDuration().seconds, LSDuration().microseconds
    



"""Usage: brilcalc beam [options] 

options:
  -h,  --help                  Show this screen.
  -c CONNECT                   DB Service name [default: offline]
  -p AUTHPATH                  Authentication file
  -n SCALEFACTOR               Scale factor on results [default: 1.0]
  -f FILL                      Fill number
  -r RUN                       Run number
  -i INPUTFILE                 Input selection file
  -o OUTPUTFILE                Output csv file. Special file '-' for stdout.
  -b BEAMSTATUS                Beam mode. FLAT TOP,SQUEEZE,ADJUST,STABLE BEAMS
  --amodetag AMODETAG          Accelerator mode 
  --beamenergy BEAMENERGY      Target single beam energy in GeV  
  --datatag DATATAG            Data tag name
  --begin BEGIN                Min start time/fill/run
  --end END                    Max start time/fill/run
  --output-style OSTYLE        Screen output style. tab, html, csv [default: tab]
  --xing                       Show result in bx granularity
  --cerntime                   Show time in CERN local time
  --tssec                      Show time as second since Epoch
 
"""

import os
from docopt import docopt
from schema import Schema
from brilws.cli import clicommonargs

def validate(optdict):
    result={}
    argdict = clicommonargs.argvalidators
    #extract sub argdict here
    myvalidables = ['-c','-n','-f','-r','-i','-o','--amodetag','-b','--beamenergy','--datatag','--begin','--end','--output-style',str]
    argdict = dict((k,v) for k,v in clicommonargs.argvalidators.iteritems() if k in myvalidables)
    schema = Schema(argdict)
    result = schema.validate(optdict)
    return result

if __name__ == '__main__':
    print(docopt(__doc__))



import numpy as np
def bxlumi(x):
    return '%d %.4f %.4f'%(x[0],x[1],x[2])
def bxintensity(x):
    return '%d %.4e %.4e'%(x[0],x[1],x[2])

def bitprescFormatter(x):
    bitname = x[0]
    prescval = x[1]
    return '%s/%d'%(bitname,prescval)

lumiunit_to_scalefactor={    
    '/kb':1.0e-09,
    '1e21/cm2':1.0e-09,
    'hz/kb':23.31*1.0e-09,
    '1e21/cm2s':23.31*1.0e-09,
    
    '/b':1.0e-06,
    '1e24/cm2':1.0e-06,
    'hz/b':23.31*1.0e-06,
    '1e24/cm2s':23.31*1.0e-06,
    
    '/mb':1.0e-03,
    '1e27/cm2':1.0e-03,
    'hz/mb':23.31*1.0e-03,
    '1e27/cm2s':23.31*1.0e-03,
    
    '/ub':1.0,
    '1e30/cm2':1.0,
    'hz/ub':23.31,
    '1e30/cm2s':23.31,
    
    '/nb':1.0e3,
    '1e33/cm2':1.0e3,
    'hz/nb':23.31*1.0e3,
    '1e33/cm2s':23.31*1.0e3,
    
    '/pb':1.0e6,
    '1e36/cm2':1.0e6,
    'hz/nb':23.31*1.0e6,
    '1e36/cm2s':23.31*1.0e6,
    
    '/fb':1.0e9,
    '1e39/cm2':1.0e9,
    'hz/fb':23.31*1.0e9,
    '1e39/cm2s':23.31*1.0e9,
    
    '/ab':1.0e12,
    '1e42/cm2':1.0e12,
    'hz/ab':23.31*1.0e12,
    '1e42/cm2s':23.31*1.0e12,
}

def lumiunit(x,unitstr):
    if x.find('(/ub)')==-1: return x
    return x.replace('(/ub)','(%s)'%unitstr)
    
#def lumiunit(x,scalefactor):
#    '''
#    return lumiunit according to scalefactor 
#    if x does not contain string /ub, do nothing
#    special if scalefactor is about 23, return hz/ub
#    else:
#        if scalefactor>=1.0e-03 and scalefactor<1, replace /ub with /mb
#        if scalefactor>=1 and scalefactor<1e3, keep /ub
#        if scalefactor>=1e3 and scalefactor<1e6, replace /ub with /nb
#        if scalefactor>=1e6 and scalefactor<1e9, replace /ub with /pb
#        if scalefactor>=1e9 and scalefactor<1e12, replace /ub with /fb
#        if scalefactor>=1e12 and scalefactor<1e15, replace /ub with /ab
#    '''
    
#    if x.find('(/ub)')==-1: return x    
#    if int(scalefactor) == 23: return x.replace('(/ub)','(hz/ub)')
#    if (scalefactor>=1.0e-09) and (scalefactor<1.0e-06): return x.replace('(/ub)','(/kb)')
#    if (scalefactor>=1.0e-06) and (scalefactor<1.0e-03): return x.replace('(/ub)','(/b)')
#    if (scalefactor>=1.0e-03) and (scalefactor<1.0): return x.replace('(/ub)','(/mb)')
#    if (scalefactor>=1.0) and (scalefactor<1.0e3): return x
#    if (scalefactor>=1.0e3) and (scalefactor<1.0e6): return x.replace('(/ub)','(/nb)')
#    if (scalefactor>=1.0e6) and (scalefactor<1.0e9): return x.replace('(/ub)','(/pb)')
#    if (scalefactor>=1.0e9) and (scalefactor<1.0e12): return x.replace('(/ub)','(/fb)')
#    if (scalefactor>=1.0e12) and (scalefactor<1.0e15): return x.replace('(/ub)','(/ab)')
#    return x

def guesslumiUnit(inverseubval):
    '''
    input:
        float value in 1/ub
    output:
        printable value (value(float),unit(str)) unit in [1/kb,1/b,1/mb,1/ub,1/nb,1/pb,1/fb]
    '''
    denomitor = 1.0
    if inverseubval>=1.0e-09 and inverseubval<1.0e-06:
        denomitor=1.0e-09
        unitstring='/kb'
        return (float(inverseubval)/float(denomitor),unitstring)
    if inverseubval>=1.0e-06 and inverseubval<1.0e-03:
        denomitor=1.0e-06
        unitstring='/b'
        return (float(inverseubval)/float(denomitor),unitstring)
    if inverseubval>=1.0e-03 and inverseubval<1.0:
        denomitor=1.0e-03
        unitstring='/mb'
        return (float(inverseubval)/float(denomitor),unitstring)
    if inverseubval>=1.0 and inverseubval<1.0e3:
        unitstring='/ub'
        return (inverseubval,unitstring)
    if inverseubval>=1.0e3 and inverseubval<1.0e06:
        denomitor=1.0e3
        unitstring='/nb'
        return (float(inverseubval)/float(denomitor),unitstring)
    if inverseubval>=1.0e6 and inverseubval<1.0e9:
        denomitor=1.0e6
        unitstring='/pb'
        return (float(inverseubval)/float(denomitor),unitstring)
    if inverseubval>=1.0e9 and inverseubval<1.0e12:
        denomitor=1.0e9
        unitstring='/fb'
        return (float(inverseubval)/float(denomitor),unitstring)
    if inverseubval>=1.0e12 and inverseubval<1.0e15:
        denomitor=1.0e12
        unitstring='/ab'
        return (float(inverseubval)/float(denomitor),unitstring)
    return (float(inverseubval),denomitor,'/ub')






import sys
import os
sys.path.insert(0,os.path.dirname(sys.executable)+'/../lib/python2.7/site-packages/') #ignore other PYTHONPATH

import logging
import docopt
import schema
import csv
import numpy as np
import pandas as pd
import brilws
import prettytable
from sqlalchemy import *
from brilws import api,params,display
from brilws.cli import clicommonargs

log = logging.getLogger('brilws')
#logformatter = logging.Formatter('%(levelname)s %(message)s')
log.setLevel(logging.ERROR)
#logging.StreamHandler().setFormatter(logformatter)
#ch = logging.StreamHandler()
#ch.setFormatter(logformatter)
#log.addHandler(ch)

def query_creationutc():
    return '''select to_char(sys_extract_utc(systimestamp), 'YY/MM/DD HH24:MI:SS') from dual'''

def briltag_main(progname=sys.argv[0]):
    
    docstr='''

    usage:
      briltag (-h|--help|--version) 
      briltag [--debug|--warn] <command> [<args>...]

    commands:
      listdata    list data tags
      listiov     list norm tags 
      insertiov   insert norm tag
      insertdata  insert data tag
    See 'briltag <command> --help' for more information on a specific command.

    '''
    argv = sys.argv[1:]
    args = docopt.docopt(docstr,argv,help=True,version=brilws.__version__,options_first=True)
    if args['--debug']:
        log.setLevel(logging.DEBUG)
    elif args['--warn']:
        log.setLevel(logging.WARNING)    
    log.debug('global arguments: %s',args)   
    cmmdargv = [args['<command>']] + args['<args>']
    
    log.debug('command arguments: %s',cmmdargv)
    parseresult = {}

    try:      
      if args['<command>'] == 'listdata':
         import briltag_listdata
         parseresult = docopt.docopt(briltag_listdata.__doc__,argv=cmmdargv)
         parseresult = briltag_listdata.validate(parseresult)

         ##parse selection params
         pargs = clicommonargs.parser(parseresult)
         dbschema = ''
         if not pargs.dbconnect.find('oracle')!=-1: dbschema = 'cms_lumi_prod'
         dbengine = create_engine(pargs.connecturl)          
         tags = api.data_gettags(dbengine,schemaname=dbschema)
         header = ['datatagnameid','datatagname','creationutc','comments']
         ptable = display.create_table(header,header=True)
         for tid,tval in tags.items():
             display.add_row( [str(tid),tval[0],tval[1],tval[2]], ptable=ptable )
         display.show_table(ptable,'tab')

      elif args['<command>'] == 'listiov':
         import briltag_listiov
         parseresult = docopt.docopt(briltag_listiov.__doc__,argv=cmmdargv)
         parseresult = briltag_listiov.validate(parseresult)
         pargs = clicommonargs.parser(parseresult)
         dbschema = ''
         if not pargs.dbconnect.find('oracle')!=-1: dbschema = 'cms_lumi_prod'
         dbengine = create_engine(pargs.connecturl)
         istypedefault = False
         if parseresult.has_key('--isdefault'):
             istypedefault = True
         if not pargs.name:            
             iovtaglist = api.iov_gettags(dbengine,datasource=pargs.lumitype,applyto=pargs.applyto,isdefault=istypedefault,schemaname=dbschema)
             header = ['tagname','tagid','creationutc','applyto','datasource','isdefault','comments']
             ptable = display.create_table(header,header=True)
             for tagname,tagcontents in iovtaglist.items():
                 display.add_row( [tagname,str(tagcontents[0]),tagcontents[1],tagcontents[2],tagcontents[3],tagcontents[4],tagcontents[5]], ptable=ptable )
             display.show_table(ptable,'tab')
         else:
             iovtagdata = api.iov_gettagdata(dbengine,pargs.name,schemaname=dbschema)
             #print iovtagdata
             header = ['since','func','payload','comments']
             ptable = display.create_table(header,header=True)
             for sincedata in iovtagdata:
                 since = sincedata[0]
                 func = sincedata[1]                 
                 payload = sincedata[2]
                 comments = sincedata[3]                
                 display.add_row( [ '%d'%since, func, payload, comments], ptable=ptable )
             display.show_table(ptable,'tab')
             
      elif args['<command>'] == 'insertiov':
         import briltag_insertiov
         parseresult = docopt.docopt(briltag_insertiov.__doc__,argv=cmmdargv)
         parseresult = briltag_insertiov.validate(parseresult)
         pargs = clicommonargs.parser(parseresult)
         dbschema = ''
         if not pargs.dbconnect.find('oracle')!=-1: dbschema = 'cms_lumi_prod'         
         dbengine = create_engine(pargs.connecturl)
         istypedefault=False
         if pargs.yamlobj.has_key('istypedefault'):
             istypedefault =  pargs.yamlobj['istypedefault']
         iovtagname = ''
         if pargs.yamlobj.has_key('name'):
             iovtagname = pargs.yamlobj['name']
         else:
             ValueError('name cannot be empty')
         applyto = 'lumi'
         if pargs.yamlobj.has_key('applyto'):
             applyto = pargs.yamlobj['applyto']
         datasource = None
         if pargs.yamlobj.has_key('datasource'):
             datasource = pargs.yamlobj['datasource']
         else:
             raise ValueError('datasource cannot be empty')
         comments = ''
         if pargs.yamlobj.has_key('comments'):
             comments = pargs.yamlobj['comments']
         istypedefault = False
         if pargs.yamlobj.has_key('istypedefault'):
             istypedefault = True
         iovdata = None
         if pargs.yamlobj.has_key('since'):
             iovdata = pargs.yamlobj['since']
         else:
             raise ValueError('since cannot be empty')
         #print iovtagname,applyto,datasource,comments,istypedefault
         iovtagid = api.iov_insertdata(dbengine,iovtagname,datasource,iovdata,applyto=applyto,isdefault=istypedefault,comments=comments,schemaname=dbschema)
                  
      elif args['<command>'] == 'insertdata':
          import briltag_insertdata
          parseresult = docopt.docopt(briltag_insertdata.__doc__,argv=cmmdargv)
          parseresult = briltag_insertdata.validate(parseresult)
          pargs = clicommonargs.parser(parseresult)
          dbschema = ''
          if not pargs.dbconnect.find('oracle')!=-1: dbschema = 'cms_lumi_prod'
          dbengine = create_engine(pargs.connecturl)
          name = pargs.name
          if not name: raise NameError('--name cannot be empty')
          datatagnameid = api.data_createtag(dbengine,datatagname=name,comments=pargs.comments,schemaname=dbschema)
          print 'created datatag %s , datatagnameid %ul , comments %s'%(name,datatagnameid,pargs.comments)     
      else:
          exit("%r is not a briltag command. See 'briltag --help'."%args['<command>']) 
    except docopt.DocoptExit:
      raise docopt.DocoptExit('Error: incorrect input format for '+args['<command>'])            
    except schema.SchemaError as e:
      exit(e)

if __name__ == '__main__':
    briltag_main()



import sys
import os
sys.path.insert(0,os.path.dirname(sys.executable)+'/../lib/python2.7/site-packages/') #ignore other PYTHONPATH

import logging
import docopt
import schema
import brilws
import prettytable
import pandas as pd
import numpy as np
import ast
from brilws import api,params,display,formatter,lumiParameters,corrector
from brilws.cli import clicommonargs
import re,time, csv
from datetime import datetime
from sqlalchemy import *
import math
from itertools import izip_longest
from dateutil import tz
import pytz
log = logging.getLogger('brilws')
logformatter = logging.Formatter('%(levelname)s %(name)s %(message)s')
log.setLevel(logging.ERROR)
ch = logging.StreamHandler()
ch.setFormatter(logformatter)
log.addHandler(ch)

lumip = lumiParameters.ParametersObject()
lslengthsec= lumip.lslengthsec()
utctmzone = tz.gettz('UTC')
cerntmzone = tz.gettz('CEST')

class Unbuffered(object):
    def __init__(self, stream):
        self.stream = stream
    def write(self, data):
        self.stream.write(data)
        self.stream.flush()
    def __getattr__(self, attr):
        return getattr(self.stream.attr)
sys.stdout = Unbuffered(sys.stdout)

np.seterr(divide='ignore', invalid='ignore')

def xing_indexfilter(arr,constfactor=1.,xingMin=0.,xingTr=0.,xingId=[]):
    vidx = None    
    if xingTr:
        vidx = np.argwhere( np.logical_and( (arr*constfactor)>xingMin, arr>xingTr*np.max(arr) ) ).ravel()
    else:
        vidx = np.argwhere( (arr*constfactor)>xingMin ).ravel()
    if xingId :
        posidx = np.array(xingId)-1 #array position index is bunch index - 1
        return np.intersect1d(np.array(posidx),vidx)    
    return vidx

def totalprescaleNEW(hltprescval,l1seedlogic,l1prescvals):
    '''
    1) if ONE single seed, take it
    2) if OR:  (A.1 OR A.2 OR A.3 ...) ignore the triggers with prescale 0, and take the minimum of the remaining prescale values:
        if max(A.i) == 0
                A = 0
        else
                A = min(A.i | A.i > 0)
    3) if AND: A AND B AND C ... take the product of the prescale values A * B * C ...
    4) (optional)
       if (A.1 OR A.2 OR A.3 ...) AND (B.1 OR B.2 OR ...) AND ...
       - for each group, take the lowest non-zero prescale, or 0 if they are all zero
         if max(A.i) = 0
                A =
         else
                A = min(A.i | A.i > 0)
       - take the product of the prescales thus obtained:
         P = A * B * ...
       users still take their own risk and should use their own judgement.
    5)  if NOT, or a combination of AND/OR and NOT, reject
    '''
    totpresc = 0
    if not hltprescval or not l1prescvals or not l1seedlogic:
        return totpresc
    if np.all( np.array(l1prescvals)==1 ):
        totpresc = hltprescval        
    elif l1seedlogic=='ONE':
        totpresc = hltprescval*l1prescvals[0]
    elif l1seedlogic=='OR':
        if np.max( np.array(l1prescvals) ) == 0:
            return totpresc
        else:
            l = np.array(l1prescvals)
            l1p = np.min( l[l>0] )
        totpresc = hltprescval*l1p
    elif l1seedlogic=='AND':
        if np.max( np.array(l1prescvals) ) == 0:
            return totpresc
        l1p = np.prod( np.array(l1prescvals) )
        totpresc = hltprescval*l1p    
    return totpresc

    
def totalprescale(hltprescval,l1seedlogic,l1prescvals):
    totpresc = 0
    if not hltprescval or not l1prescvals or not l1seedlogic:
        return totpresc
    if np.all( np.array(l1prescvals)==1 ):
        totpresc = hltprescval
    elif l1seedlogic=='ONE':
        totpresc = hltprescval*l1prescvals[0]
    elif l1seedlogic=='OR':
        
        totpresc = hltprescval*np.min(l1prescvals)
    elif l1seedlogic=='AND':
        totpresc = hltprescval*np.max(l1prescvals)    
    return totpresc

def lumi_per_normtag(shards,lumiquerytype,dbengine,dbschema,runtot,datasource=None,normtag=None,withBX=False,byls=None,fh=None,csvwriter=None,ptable=None,scalefactor=1,totz=utctmzone,fillmin=None,fillmax=None,runmin=None,runmax=None,amodetagid=None,egev=None,beamstatusid=None,tssecmin=None,tssecmax=None,runlsSeries=None,hltl1map={},ignorel1mask=False,xingMin=0.,xingTr=0.,xingId=[],checkjson=False,datatagnameid=None):

    validitychecker = None
    lastvalidity = None
    if normtag and normtag is not 'withoutcorrection':
        normdata = api.iov_gettagdata(dbengine, normtag,schemaname=dbschema)
        if not normdata: raise ValueError('normtag %s does not exist'%normtag)
        validitychecker = ValidityChecker(normdata)
        
    for shard in shards:              
        lumiiter = None
        if lumiquerytype == 'bestresultonline':
            tablename = 'online_result_'+str(shard)
            shardexists = api.table_exists(dbengine,tablename,schemaname=dbschema)
            if not shardexists: continue
            rfields = ['fillnum','runnum','lsnum','timestampsec','cmson','beamstatusid','targetegev','delivered','recorded','avgpu','datasource','numbxbeamactive']
            if withBX: rfields = rfields+['bxdeliveredblob']                    
            lumiiter = api.online_resultIter(dbengine,tablename,schemaname=dbschema,fields=rfields,fillmin=fillmin,fillmax=fillmax,runmin=runmin,runmax=runmax,amodetagid=amodetagid,targetegev=egev,beamstatusid=beamstatusid,tssecmin=tssecmin,tssecmax=tssecmax,runlsselect=runlsSeries,sorted=True)
            
        elif lumiquerytype in ['detresultonline','detraw']:
            datatype = 'raw'
            if lumiquerytype=='detresultonline':
                datatype = 'result'
                
            tablename = datasource.lower()+'_'+datatype+'_'+str(shard)
            shardexists = api.table_exists(dbengine,tablename,schemaname=dbschema)
            if not shardexists: continue
            rfields = ['avglumi']
            idfields = ['fillnum','runnum','lsnum','timestampsec','beamstatusid','cmson','deadtimefrac','targetegev','numbxbeamactive']
            if withBX: rfields = rfields+['bxlumiblob']
            lumiiter = api.dataIter(dbengine,datasource,datatype,shard,datafields=rfields,idfields=idfields,schemaname=dbschema,fillmin=fillmin,fillmax=fillmax,runmin=runmin,runmax=runmax,amodetagid=amodetagid,targetegev=egev,beamstatusid=beamstatusid,tssecmin=tssecmin,tssecmax=tssecmax,runlsselect=runlsSeries,sorted=True,datatagnameid=datatagnameid)                    
                  
        if lumiiter is None:
            continue

        g_run_old = 0
        g_ls_trglastscaled_old = 0
        prescale_map = {} # for global scope
        g_hltconfigid = 0
        for row in lumiiter:
            fillnum = row['fillnum']
            runnum = row['runnum']
            lsnum = row['lsnum']            
            cmslsnum = lsnum
            timestampsec = row['timestampsec']
            cmson = row['cmson']
            if not cmson:
                cmslsnum = 0
            if hltl1map:
                if cmslsnum==0: continue  #cms is not running, skip.                
                if g_run_old!=runnum:     #on new run boundary, get hltconfigid
                    hltrunconfig = api.get_hltrunconfig(dbengine,runnum=runnum,schemaname=dbschema) #[['runnum','hltconfigid','hltkey']]
                    if not hltrunconfig:
                        continue
                    g_hltconfigid = [h[1] for h in hltrunconfig][0]
                    if not g_hltconfigid:
                        continue      
                    presc = api.get_prescidx_change(dbengine,runnum,schemaname=dbschema)#{runnum:[[lslastscaler,prescidx]] }
                    if presc.has_key(runnum):
                        presc = presc[runnum]
                this_prescidx = None
                if not presc:                    
                    ls_trglastscaled = 1
                else:
                    b = [ i[0] for i in presc if i[0]<=cmslsnum ] 
                    if not b:
                        ls_trglastscaled = 1
                        this_prescidx = None                        
                    else:
                        ls_trglastscaled = np.max( b )
                        this_prescidx = [t[1] for t in presc if t[0]==ls_trglastscaled][0]
                if (g_ls_trglastscaled_old != ls_trglastscaled) or (g_run_old != runnum): #on prescale change lumi section or new run
                    prescale_map = {} #clear prescaleindex change map on new prescale index value
                    if not hltl1map.has_key(g_hltconfigid):
                        continue
                    this_hltl1map = hltl1map[g_hltconfigid]#[[hltpathid,hltpathname,l1seedtype,l1seedbits]]
                    for grouped in this_hltl1map:
                        this_hltpathid = grouped[0]
                        hltpathname = grouped[1]
                        l1seedlogic = grouped[2]
                        l1candidates = grouped[3]
                        if this_prescidx is None:
                            hltprescval = 1
                            totpresc = 1
                        else:
                            hltprescval = api.get_hltprescale(dbengine,runnum,ls_trglastscaled,g_hltconfigid,this_prescidx,this_hltpathid,schemaname=dbschema)
                            rl = api.get_l1prescale(dbengine,runnum,ls_trglastscaled,l1candidates=l1candidates,prescidxs=this_prescidx,ignorel1mask=ignorel1mask ,schemaname=dbschema)
                            if not rl: continue
                            l1keys = [ k for k in rl.keys() if k[0]==this_prescidx ]
                            l1vals = [ rl[k] for k in l1keys ]                       
                            l1prescvals = [ v[0] for v in l1vals ]                            
                            totpresc = totalprescaleNEW(hltprescval,l1seedlogic,l1prescvals)
                            if not totpresc:
                                continue
                        prescale_map[hltpathname] = totpresc   
                    g_ls_trglastscaled_old = ls_trglastscaled
                g_run_old = runnum
            beamstatusid = row['beamstatusid']
            beamstatus = params._idtobeamstatus[beamstatusid]
            if beamstatus not in ['FLAT TOP','STABLE BEAMS','SQUEEZE','ADJUST']: continue
            tegev = row['targetegev']
            dtime = str(timestampsec)
            if totz is not None:
                d = datetime.fromtimestamp(int(timestampsec),tz=pytz.utc)
                dtime = d.astimezone(totz).strftime(params._datetimefm)
            delivered = recorded = avgpu = livefrac = 0.
            if checkjson:
                g_returnedls.append((runnum,lsnum))
            if lumiquerytype == 'bestresultonline': ##bestlumi
                if row.has_key('delivered') and row['delivered']:                    
                    delivered = np.true_divide(row['delivered']*lslengthsec,scalefactor)
                if delivered>0 and row.has_key('recorded') and row['recorded']:
                    recorded = np.true_divide(row['recorded']*lslengthsec,scalefactor)
                if delivered>0 and row.has_key('avgpu') and row['avgpu']:
                    avgpu = row['avgpu']
                ds = 'UNKNOWN' 
                if row.has_key('datasource') and row['datasource']:
                    ds = row['datasource']
                livefrac = np.true_divide(recorded,delivered)
                if withBX:    #--xing
                    bxlumi = None
                    bxlumistr = '[]'
                    if row.has_key('bxdeliveredblob'):
                        bxdeliveredarray = np.array(api.unpackBlobtoArray(row['bxdeliveredblob'],'f'))                        
                        totfactor = np.true_divide(lslengthsec,scalefactor)
                        bxidx = xing_indexfilter(bxdeliveredarray,constfactor=totfactor,xingMin=xingMin,xingTr=xingTr,xingId=xingId)
                        if bxidx is not None and bxidx.size>0:
                            bxdelivered = bxdeliveredarray[bxidx]*totfactor
                            bxlumi = np.transpose( np.array([bxidx+1,bxdelivered,bxdelivered*livefrac]) )
                        del bxdeliveredarray
                        del bxidx
                        if hltl1map: #--hltpath bx display
                            for pth in prescale_map.keys():                                
                                thispresc = prescale_map[pth]
                                if bxlumi is not None:
                                    a = map(formatter.bxlumi,bxlumi/thispresc)  
                                    bxlumistr = '['+' '.join(a)+']'                                
                                display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,pth,'%.3f'%(np.true_divide(delivered,thispresc)),'%.3f'%(np.true_divide(recorded,thispresc)),ds,'%s'%bxlumistr] , fh=fh, csvwriter=csvwriter, ptable=ptable)                                
                        else:        #normal bx display                   
                            if bxlumi is not None:
                                a = map(formatter.bxlumi,bxlumi)
                                bxlumistr = '['+' '.join(a)+']'
                            display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,beamstatus,'%d'%tegev,'%.3f'%(delivered),'%.3f'%(recorded),'%.1f'%(avgpu),ds,'%s'%bxlumistr] , fh=fh, csvwriter=csvwriter, ptable=ptable)                            
                    del bxlumi
                    
                elif byls:      #--byls
                    if hltl1map:#--hltpath display
                        for pth in prescale_map.keys():
                            thispresc = prescale_map[pth]
                            display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum), dtime, pth, '%.3f'%(np.true_divide(delivered,thispresc)),'%.3f'%(np.true_divide(recorded,thispresc)),'%.1f'%(avgpu),ds] , fh=fh, csvwriter=csvwriter, ptable=ptable)                            
                    else:       #normal display
                        display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,beamstatus,'%d'%tegev,'%.3f'%(delivered),'%.3f'%(recorded),'%.1f'%(avgpu),ds] , fh=fh, csvwriter=csvwriter, ptable=ptable)                        
            else:               ###lumisource
                if row.has_key('deadtimefrac') :
                    if row['deadtimefrac'] is not None:
                        livefrac = 1.-row['deadtimefrac']
                    else:
                        if cmslsnum!=0:
                            g_nulldeadtime.setdefault(runnum,[]).append(lsnum)
                        livefrac = 0.
                uncorrectedavglumi = row['avglumi'] #uncorrected, hz/ub
                avglumi = uncorrectedavglumi
                ncollidingbx = row['numbxbeamactive']
                uncorrectedbxlumi = None
                if row.has_key('bxlumiblob'):
                    uncorrectedbxlumi = np.array(api.unpackBlobtoArray(row['bxlumiblob'],'f'))
                bxlumi = uncorrectedbxlumi
                if validitychecker is not None:
                    if not lastvalidity or not validitychecker.isvalid(runnum,lastvalidity):
                        lastvalidity = validitychecker.getvalidity(runnum)
                    [normfunc,normparam] = validitychecker.getvaliddata(lastvalidity[0])
                    if withBX:
                        (avglumi,bxlumi) = corrector.applyCorrectionStr( normfunc,normparam, corrector.FunctionRoot(uncorrectedbxlumi,avglumi,ncollidingbx) )
                    else:
                        avglumi = corrector.applyCorrectionStr( normfunc,normparam, corrector.FunctionRoot(avglumi,avglumi,ncollidingbx) )[0]

                totfactor = np.true_divide(lslengthsec,scalefactor)
                delivered = avglumi*totfactor
                recorded = delivered*livefrac
                avgpu = lumip.avgpu( avglumi,ncollidingbx,g_minbias)                
                if withBX:      #--xing
                    bxlumistr = '[]'
                    if bxlumi is not None:                                                         
                        bxidx = xing_indexfilter(bxlumi,constfactor=totfactor,xingMin=xingMin,xingTr=xingTr,xingId=xingId)
                        if bxidx is not None and bxidx.size>0:                                      
                            bxdelivered =  bxlumi[bxidx]*totfactor
                            bxlumi = np.transpose( np.array([bxidx+1,bxdelivered,bxdelivered*livefrac]) )               
                        del uncorrectedbxlumi
                        del bxidx
                    if hltl1map:#--hltpath xing display
                        for pth in prescale_map.keys():
                            thispresc = prescale_map[pth]                            
                            if bxlumi is not None:
                                a = map(formatter.bxlumi,bxlumi/thispresc)  
                                bxlumistr = '['+' '.join(a)+']'                                
                            display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,pth,'%.3f'%(np.true_divide(delivered,thispresc)),'%.3f'%(np.true_divide(recorded,thispresc)),datasource.upper(),'%s'%bxlumistr] , fh=fh, csvwriter=csvwriter, ptable=ptable)                            
                    else:       #normal xing display
                        if bxlumi is not None:
                            a = map(formatter.bxlumi,bxlumi)
                            bxlumistr = '['+' '.join(a)+']'
                        display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,beamstatus,'%d'%tegev,'%.3f'%(delivered),'%.3f'%(recorded),'%.1f'%(avgpu),datasource.upper(),'%s'%bxlumistr] , fh=fh, csvwriter=csvwriter, ptable=ptable)                        
                    del bxlumi
                elif byls:       #--byls
                    if hltl1map: #--hltpath display
                        for pth in prescale_map.keys():
                            thispresc = prescale_map[pth]                            
                            display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,pth,'%.3f'%(np.true_divide(delivered,thispresc)),'%.3f'%(np.true_divide(recorded,thispresc)),'%.1f'%(avgpu),datasource.upper()] , fh=fh, csvwriter=csvwriter, ptable=ptable)                            
                    else:        #normal display
                        display.add_row( ['%d:%d'%(runnum,fillnum),'%d:%d'%(lsnum,cmslsnum),dtime,beamstatus,'%d'%tegev,'%.3f'%(delivered),'%.3f'%(recorded),'%.1f'%(avgpu),datasource.upper()] , fh=fh, csvwriter=csvwriter, ptable=ptable)
            if not hltl1map:     #normal statistic collect
                if not runtot.has_key( ('',runnum) ):
                    runtot[('',runnum)] = {'fill':fillnum,'dtime':dtime,'nls':0,'ncms':0,'delivered':0,'recorded':0} 
                runtot[ ('',runnum)]['nls']+=1                
                if cmson: runtot[ ('',runnum) ]['ncms']+=1
                runtot[ ('',runnum) ]['delivered']+=delivered
                runtot[ ('',runnum) ]['recorded']+=recorded                                
            else:                #hltpath statistic collect
                if not cmson:
                    continue
                for pth in prescale_map.keys():
                    thispresc = prescale_map[pth]
                    if not runtot.has_key( (pth,runnum) ):
                        runtot[(pth,runnum)] = {'fill':fillnum,'dtime':dtime,'nls':0,'ncms':0,'delivered':0,'recorded':0}                        
                    runtot[ (pth,runnum)]['nls']+=1                
                    if cmson: runtot[ (pth,runnum) ]['ncms'] += 1
                    runtot[ (pth,runnum) ]['delivered'] += np.true_divide(delivered,thispresc)
                    runtot[ (pth,runnum) ]['recorded'] += np.true_divide(recorded,thispresc)      
                
class ValidityChecker(object):
    def __init__(self, normdata):
        self.normdata = normdata
        self.allsince = np.array([x[0] for x in normdata])
        
    def getvalidity(self,runnum):
        if self.allsince.size == 0 : return None
        since = 1
        if self.allsince[self.allsince<=runnum].size>0:
            since = self.allsince[self.allsince<=runnum].max()
        till = 999999
        if self.allsince[self.allsince>runnum].size>0:
            till = self.allsince[self.allsince>runnum].min()
        return (since,till)
    
    def isvalid(self,runnum,validity):
        if runnum>=validity[0] and runnum<validity[1]:
            return True
        return False

    def getvaliddata(self,since):
        s = np.where(self.allsince==since)
        if s[0].size>0:
            sinceindex = s[0][0]
            func = self.normdata[sinceindex][1]
            params = self.normdata[sinceindex][2]
            return [func,params]
        return None

def findtagname(dbengine,datatagname,dbschema):
    '''
    output: (datatagname,datatagnameid)
    '''
    datatagnameid=0
    if datatagname:
        datatagnameid = api.datatagnameid(dbengine,datatagname=datatagname,schemaname=dbschema)
    else:
        r = api.max_datatagname(dbengine,schemaname=dbschema)
        if not r:
            raise RuntimeError('no tag found')
        datatagname = r[0]
        datatagnameid = r[1]
    if not datatagnameid:
        raise RuntimeError('No tag %s found'%datatagname)   
    return (datatagname,datatagnameid)

          
def brilcalc_main(progname=sys.argv[0]):

    docstr='''

    usage:
      brilcalc (-h|--help|--version) 
      brilcalc [--debug|--warn] <command> [<args>...]

    commands:
      lumi Luminosity
      beam Beam 
      trg Trigger configurations      
    See 'brilcalc <command> --help' for more information on a specific command.

    '''
    args = {}
    argv = sys.argv[1:]
    args = docopt.docopt(docstr,argv,help=True,version=brilws.__version__,options_first=True)

    if args['--debug']:
        log.setLevel(logging.DEBUG)
    elif args['--warn']:
        log.setLevel(logging.WARNING)
        
    log.debug('global arguments: %s',args)
    cmmdargv = [args['<command>']] + args['<args>']

    log.debug('command arguments: %s',cmmdargv)
    parseresult = {}
    global g_returnedls #[(run,ls),...]
    global g_nulldeadtime #{run:[]}
    global g_minbias #minbias xsec
    try:
      if args['<command>'] == 'lumi':
          import brilcalc_lumi          
          parseresult = docopt.docopt(brilcalc_lumi.__doc__,argv=cmmdargv)
          parseresult = brilcalc_lumi.validate(parseresult)          
          ##parse selection params
          pargs = clicommonargs.parser(parseresult)
          dbschema = ''
          if not pargs.dbconnect.find('oracle')!=-1: dbschema = 'cms_lumi_prod'
          log.debug('connecturl: %s'%pargs.connecturl)
          dbengine = create_engine(pargs.connecturl)
          
          selectionkwds = {}
          normtag = None          
                      
          fh = None
          ptable = None
          ftable = None
          csvwriter = None
          vfunc_lumiunit = np.vectorize(formatter.lumiunit)
          checkjson = True
          if parseresult['--without-checkjson'] or not parseresult['-i']:
              checkjson = False
              
          g_returnedls = []
          g_nulldeadtime = {}
          g_headers = {}
          g_headers['runheader'] = ['run:fill','time','nls','ncms','delivered(/ub)','recorded(/ub)']
          g_headers['footer'] = ['nfill','nrun','nls','ncms','totdelivered(/ub)','totrecorded(/ub)']
          g_headers['bylsheader'] = ['run:fill','ls','time','beamstatus','E(GeV)','delivered(/ub)','recorded(/ub)','avgpu','source']
          g_headers['runheader_hltpath'] = ['run:fill','time','ncms','hltpath','delivered(/ub)','recorded(/ub)']
          g_headers['footer_hltpath'] = ['hltpath','nfill','nrun','ncms','totdelivered(/ub)','totrecorded(/ub)']
          g_headers['bylsheader_hltpath'] = ['run:fill','ls','time','hltpath','delivered(/ub)','recorded(/ub)','avgpu','source']
          g_minbias = pargs.minbias
          scalefactor = pargs.scalefactor
          lumiunitstr = parseresult['-u']         
          if lumiunitstr not in formatter.lumiunit_to_scalefactor.keys(): raise ValueError('%s not recognised as lumi unit'%lumiunit)
          lumiunitconversion = formatter.lumiunit_to_scalefactor[lumiunitstr]
          scalefactor = scalefactor*lumiunitconversion

          if pargs.withBX:
              if pargs.hltpath is None:
                  header = g_headers['bylsheader']+['[bxidx bxdelivered(/ub) bxrecorded(/ub)]']
                  footer = g_headers['footer']
              else:
                  header = g_headers['bylsheader_hltpath']+['[bxidx bxdelivered(/ub) bxrecorded(/ub)]']
                  footer = g_headers['footer_hltpath']
          elif pargs.byls:
              if pargs.hltpath is None:
                  header = g_headers['bylsheader']
                  footer = g_headers['footer']
              else:
                  header = g_headers['bylsheader_hltpath']
                  footer = g_headers['footer_hltpath']
          else:
              if pargs.hltpath is None:
                  header = g_headers['runheader']
                  footer = g_headers['footer']
              else:
                  header = g_headers['runheader_hltpath']
                  footer = g_headers['footer_hltpath']
                  
          header = vfunc_lumiunit(header,lumiunitstr).tolist()
          footer = vfunc_lumiunit(footer,lumiunitstr).tolist()
        
          
          (datatagname,datatagnameid) = findtagname(dbengine,pargs.datatagname,dbschema)
          if not pargs.totable:
              fh = pargs.ofilehandle
              print >> fh, '#Data tag : %s , Norm tag: %s'%(datatagname,normtag)
              print >> fh, '#'+','.join(header)
              csvwriter = csv.writer(fh)
          else:
              ptable = display.create_table(header,header=True,align='l')
              ftable = display.create_table(footer,header=True,align='l')          
              
          datasources = [] #[lumiquerytype,normtagname,datasource,runlsstr]          

          lumiquerytype = 'detraw'
          normtag = normtagname = 'withoutcorrection'
          parsediffruns = None
          parsediffls = None
          if not pargs.withoutcorrection:
              normtag = pargs.iovtagSelect
              if not normtag:                  
                  if pargs.lumitype:
                      lumiquerytype = 'detresultonline'
                      datasources.append( ['detresultonline',normtag,pargs.lumitype.lower(),pargs.runlsSeries] )
                      normtagname = 'onlineresult'
                  else:
                      lumiquerytype = 'bestresultonline'
                      normtagname = 'onlineresult'
                      datasources.append( ['bestresultonline',normtag,'best',pargs.runlsSeries] )
              else:
                  if isinstance(normtag,list): #normtag is list
                      if pargs.runlsSeries is None:
                          mergedselect = normtag
                      else:
                          (parsediffruns, parsediffls) = api.checksuperset([x[1] for x in normtag],pargs.runlsSeries)                         
                          mergedselect = api.mergeiovrunls(normtag,pargs.runlsSeries)
                      normtagname = 'composite'
                      for item in mergedselect:
                          iovtag = item[0]
                          runlsdict = item[1]
                          iovtag_meta = api.iov_gettag(dbengine,iovtag,schemaname=dbschema)
                          if not iovtag_meta: raise ValueError('requested iovtags do not exist')  
                          datasource = iovtag_meta[2].lower()
                          datasources.append( [lumiquerytype,iovtag,datasource,runlsdict] )
                  else:  #normtag is string 
                      datasource = pargs.lumitype
                      normtagname = normtag
                      if datasource is None: #det lumi with correction
                          iovtag_meta = api.iov_gettag(dbengine,normtag,schemaname=dbschema)
                          if not iovtag_meta: raise ValueError('%s does not exist'%normtag)                   
                          datasource = iovtag_meta[2]
                      datasources.append( [lumiquerytype,normtag,datasource.lower(),pargs.runlsSeries ])
                     
          else:
              if not pargs.lumitype:
                  raise ValueError('--type is required with --without-correction')
              datasources.append( [lumiquerytype,None,pargs.lumitype.lower(),pargs.runlsSeries ])
          
          log.debug('lumiquerytype %s'%lumiquerytype)          
          log.debug('scalefactor: %.2f'%pargs.scalefactor)                    
          log.debug('minbias: %.1f'%g_minbias)
          
          runtot = {} #{(hltpath,run): { 'fill':fillnum,'time':dtime,'nls':1,'ncms':int(cmson),'delivered':delivered,'recorded':recorded} }
                               
          totz=utctmzone
          if pargs.cerntime:
              totz=cerntmzone
          elif pargs.tssec:
              totz=None          

          fillmin = pargs.fillmin
          fillmax = pargs.fillmax
          runmin = pargs.runmin
          runmax = pargs.runmax
          amodetagid = pargs.amodetagid
          egev = pargs.egev
          beamstatusid = pargs.beamstatusid
          tssecmin = pargs.tssecmin
          tssecmax = pargs.tssecmax
          
          hltl1map = None
          
          if pargs.hltpath is not None:
              hltl1map = api.get_hlttrgl1seedmap(dbengine,hltpath=pargs.hltpath,schemaname=dbschema)
              if not hltl1map:
                  print 'no hltpath to l1bit mapping found'
                  sys.exit(0)
          
          rselectrange = [] 
          if datasources :
              for [qt,nt,ds,rs] in datasources:
                  if rs is None: continue
                  for r,l in rs.iteritems():
                      if not r in rselectrange:
                          rselectrange.append(r)              
          shards = api.locate_shards(dbengine,runmin=runmin,runmax=runmax,fillmin=fillmin,fillmax=fillmax,tssecmin=tssecmin,tssecmax=tssecmax,orrunlist=rselectrange,schemaname=dbschema)
          if not shards:
              print 'Failed to find data table for the requested time range.'
              sys.exit(1)
          log.debug('shards: '+str(shards))
          
          for [qtype,ntag,dsource,rselect] in datasources:             
              lumi_per_normtag(shards,qtype,dbengine,dbschema,runtot,datasource=dsource,normtag=ntag,withBX=pargs.withBX,byls=pargs.byls,fh=fh,csvwriter=csvwriter,ptable=ptable,scalefactor=scalefactor,totz=totz,fillmin=fillmin,fillmax=fillmax,runmin=runmin,runmax=runmax,amodetagid=amodetagid,egev=egev,beamstatusid=beamstatusid,tssecmin=tssecmin,tssecmax=tssecmax,runlsSeries=rselect,hltl1map=hltl1map,ignorel1mask=parseresult['--ignore-mask'],xingMin=pargs.xingMin,xingTr=pargs.xingTr,xingId=pargs.xingId,checkjson=checkjson,datatagnameid=datatagnameid)  
          
          if pargs.hltpath is None:
              nruns = len(runtot.keys())
              nfills = len( set([v['fill'] for v in runtot.values()] ) )
              nls = np.sum( [v['nls'] for v in runtot.values()]  )
              ncmsls = np.sum( [v['ncms'] for v in runtot.values()] )
              totdelivered = np.sum( [v['delivered'] for v in runtot.values() ] )
              totrecorded =  np.sum( [v['recorded'] for v in runtot.values() ] )
              if not pargs.byls and not pargs.withBX: #run table
                  for hn,rn in sorted(runtot.keys()):
                      v = runtot[ (hn,rn) ]
                      display.add_row( ['%d:%d'%(rn,v['fill']),v['dtime'],v['nls'],v['ncms'],'%.3f'%(v['delivered']),'%.3f'%(v['recorded']) ] , fh=fh, csvwriter=csvwriter, ptable=ptable)
              display.add_row( [ '%d'%nfills, '%d'%nruns, '%d'%nls,'%d'%ncmsls,'%.3f'%(totdelivered),'%.3f'%(totrecorded)], fh=fh, csvwriter=None, ptable=ftable)              
                  
          else:
              hltpathSummary = []
              runtot_df = pd.DataFrame.from_dict(runtot,orient='index')
              grouped = runtot_df.groupby(level=0) #group by hltpath
              for hn,g in grouped:
                  nruns = g.index.size
                  nfills = int(g['fill'].nunique())
                  ncmsls =  int(g['ncms'].sum())
                  totdelivered = g['delivered'].sum()
                  totrecorded = g['recorded'].sum()
                  display.add_row( [ hn, '%d'%nfills, '%d'%nruns, '%d'%ncmsls,'%.3f'%(totdelivered),'%.3f'%(totrecorded)], fh=fh, csvwriter=None, ptable=ftable)                  
                  if not pargs.byls and not pargs.withBX: #hltpath, run table
                      for i,v in g.iterrows():
                          pname = i[0]
                          trun = i[1]
                          tfill = v['fill']
                          ttime = v['dtime']
                          tncmsls = v['ncms']
                          tdelivered = v['delivered']
                          trecorded = v['recorded']
                          display.add_row( [ '%d:%d'%(trun,tfill), ttime, tncmsls, pname, '%.3f'%tdelivered, '%.3f'%trecorded], fh=fh, csvwriter=csvwriter, ptable=ptable )
                  hltpathSummary.append([hn,nfills,nruns,ncmsls,totdelivered,totrecorded])   
              del runtot_df
          
          if pargs.totable:              
              print '#Data tag : %s , Norm tag: %s'%(datatagname,normtagname)
              display.show_table(ptable,pargs.outputstyle)
              print "#Summary: "
              display.show_table(ftable,pargs.outputstyle)
              del ptable
              del ftable
          else:              
              print >> fh, '#Summary:'                  
              print >> fh, '#'+','.join(footer)
              if not pargs.hltpath:
                  print >> fh, '#'+','.join( [ '%d'%nfills,'%d'%nruns,'%d'%nls,'%d'%ncmsls,'%.3f'%(totdelivered),'%.3f'%(totrecorded)] )
              else:
                  for pentry in hltpathSummary:
                      print >> fh, '#'+','.join( [ '%s'%pentry[0],'%d'%pentry[1],'%d'%pentry[2],'%d'%pentry[3],'%.3f'%pentry[4],'%.3f'%pentry[5] ] )
          if parsediffruns or parsediffls:
              print '\nWarning: problems found in merging -i and --normtag selections:'
              if parsediffruns:
                  print '  runs %s are not covered by normtag'%(parsediffruns)
              if parsediffls:
                  for r in parsediffls.keys():
                      print '  in run %d %s is not a superset of %s'%(r, parsediffls[r][0], parsediffls[r][1])
          print         

          if checkjson:
              selectlist = []
              for k,v in rselect.iteritems():
                  for vv in v:
                      rlist = api.expandrange(vv)
                      for r in rlist:
                          selectlist.append((k,r))
              arr_selectlist = np.array(selectlist,dtype=np.dtype('int,int'))
              arr_returnedls = np.array(g_returnedls,dtype=np.dtype('int,int'))
              clist = list(np.setdiff1d(arr_selectlist,arr_returnedls,assume_unique=True))
              if pargs.totable:
                  print '#Check JSON:'                  
                  print '#(run,ls) in json but not in results: %s'%(str(clist))
              else:
                  print >> fh, '#Check JSON:'
                  print >> fh, '#(run,ls) in json but not in results: %s'%(str(clist))

          if g_nulldeadtime:
              if pargs.totable:
                  print '#WARN: unknown deadtime while CMS is on'   
                  for rr,lss in g_nulldeadtime.items():
                      print '#%d %s'%(rr,str(lss))
              else:
                  print >> fh, '#WARN: unknown deadtime while CMS is on'
                  for rr,lss in g_nulldeadtime.items():
                      print >> fh, '#%d %s'%(rr,str(lss))
                      
          if fh and fh is not sys.stdout: fh.close()
          sys.exit(0)

      elif args['<command>'] == 'beam':
          import brilcalc_beam

          parseresult = docopt.docopt(brilcalc_beam.__doc__,argv=cmmdargv)
          parseresult = brilcalc_beam.validate(parseresult)
          ##parse selection params
          pargs = clicommonargs.parser(parseresult)

          ##db params
          dbschema = ''
          if not pargs.dbconnect.find('oracle')!=-1: dbschema = 'cms_lumi_prod'
          dbengine = create_engine(pargs.connecturl)
          totz=utctmzone
          if pargs.cerntime: totz=cerntmzone
          if pargs.tssec: totz=None
          ##display params          
          fh = None
          ptable = None
          csvwriter = None

          (datatagname,datatagnameid) = findtagname(dbengine,pargs.datatagname,dbschema)
          log.debug('datatagname: %s, datatagnameid: %d'%(datatagname,datatagnameid))    
          header = ['fill','run','ls','time','egev','intensity1','intensity2','ncollidingbx']
          if pargs.withBX:
              header = ['fill','run','ls','time','[bxidx intensity1 intensity2]']
          if not pargs.totable:
              fh = pargs.ofilehandle
              print >> fh, '#Data tag : %s'%(datatagname)
              print >> fh, '#'+','.join(header)
              csvwriter = csv.writer(fh)
          else:
              ptable = display.create_table(header,header=True,maxwidth=80,align='l')                        

          idfields = ['fillnum','runnum','lsnum','timestampsec','beamstatusid','numbxbeamactive']    
          fields = ['egev','intensity1','intensity2']
          if pargs.withBX:
              fields = ['bxidxblob','bxintensity1blob','bxintensity2blob']
              
          rselectrange = []
          if pargs.runlsSeries is not None:
              for r,l in pargs.runlsSeries.iteritems():
                  if not r in rselectrange:
                      rselectrange.append(r)          
          shards = api.locate_shards(dbengine,runmin=pargs.runmin,runmax=pargs.runmax,fillmin=pargs.fillmin,fillmax=pargs.fillmax,tssecmin=pargs.tssecmin,tssecmax=pargs.tssecmax,orrunlist=rselectrange,schemaname=dbschema)
          if not shards:
              print 'Failed to find data table for the requested time range.'
              sys.exit(1)
          log.debug('shards: '+str(shards))
          for shard in shards:
              beamIt = api.dataIter(dbengine,'beam','',shard,datafields=fields,idfields=idfields,schemaname=dbschema,fillmin=pargs.fillmin,fillmax=pargs.fillmax,runmin=pargs.runmin,runmax=pargs.runmax,amodetagid=pargs.amodetagid,targetegev=pargs.egev,beamstatusid=pargs.beamstatusid,tssecmin=pargs.tssecmin,tssecmax=pargs.tssecmax,runlsselect=pargs.runlsSeries,sorted=True,datatagnameid=datatagnameid)
              if not beamIt:
                  continue
              for row in beamIt:
                  fillnum = row['fillnum']
                  runnum = row['runnum']
                  lsnum = row['lsnum']                          
                  timestampsec = row['timestampsec']
                  dtime = str(timestampsec)
                  ncollidingbx = 0
                  if totz is not None:
                      d = datetime.fromtimestamp(int(timestampsec),tz=pytz.utc)
                      dtime = d.astimezone(totz).strftime(params._datetimefm) 
                  if pargs.withBX:
                      bxintensity = None
                      bxintensitystr = '[]'
                      if row.has_key('bxidxblob') and row['bxidxblob'] is not None:
                          bxidxarray = np.array(api.unpackBlobtoArray(row['bxidxblob'],'H'))                            
                          bxidxarray = bxidxarray[bxidxarray!=np.array(None)]
                          if bxidxarray is not None and bxidxarray.size>0:
                              bxintensity1array =  np.array(api.unpackBlobtoArray(row['bxintensity1blob'],'f'))
                              bxintensity2array =  np.array(api.unpackBlobtoArray(row['bxintensity2blob'],'f'))
                              bxintensity = np.transpose( np.array([bxidxarray+1,bxintensity1array,bxintensity2array]) )
                              a = map(formatter.bxintensity,bxintensity)                          
                              bxintensitystr = '['+' '.join(a)+']'
                              del bxintensity1array
                              del bxintensity2array
                          del bxidxarray
                          display.add_row( ['%d'%fillnum,'%d'%runnum,'%d'%lsnum,dtime,'%s'%bxintensitystr], fh=fh, csvwriter=csvwriter, ptable=ptable )
                  else:
                      egev = row['egev']
                      intensity1 = row['intensity1']/pargs.scalefactor
                      intensity2 = row['intensity2']/pargs.scalefactor
                      ncollidingbx = row['numbxbeamactive'] 
                      display.add_row( ['%d'%fillnum,'%d'%runnum,'%d'%lsnum,dtime,'%.1f'%egev,'%.4e'%intensity1,'%.4e'%intensity2, '%d'%ncollidingbx],fh=fh, csvwriter=csvwriter, ptable=ptable)

          if pargs.totable:
              print '#Data tag : ',datatagname
              display.show_table(ptable,pargs.outputstyle)
              del ptable
                  
          if fh and fh is not sys.stdout: fh.close()
          sys.exit(0)
          
      elif args['<command>'] == 'trg':      
          import brilcalc_trg
          parseresult = docopt.docopt(brilcalc_trg.__doc__,argv=cmmdargv)
          parseresult = brilcalc_trg.validate(parseresult)
          ##parse selection params
          pargs = clicommonargs.parser(parseresult)

          ##db params
          dbschema = ''
          if not pargs.dbconnect.find('oracle')!=-1: dbschema = 'cms_lumi_prod'
          dbengine = create_engine(pargs.connecturl)
                  
          ##display params          
          fh = None
          ptable = None
          csvwriter = None

          is_pathinfo = parseresult['--pathinfo']
          is_prescale = parseresult['--prescale']
            
          if is_pathinfo:
              header = ['hltpath','logic','l1bit']
              hltrunconfig = []
              hltconfigids = []
              if pargs.hltconfigid or pargs.hltkey or pargs.runmin:
                  hltrunconfig = api.get_hltrunconfig(dbengine,hltconfigid=pargs.hltconfigid,hltkey=pargs.hltkey,runnum=pargs.runmin,schemaname=dbschema)  #[['runnum','hltconfigid','hltkey'],['runnum','hltconfigid','hltkey']]
                  if not hltrunconfig:
                      print 'hltconfig not found'
                      sys.exit(0)
                  hltconfigids = np.unique( [h[1] for h in hltrunconfig] )
                  
              hltl1map = api.get_hlttrgl1seedmap(dbengine,hltpath=pargs.hltpath,hltconfigids=hltconfigids,schemaname=dbschema)
              # {hltconfigid: [[hltpathid,hltpathname,seedtype,[seedvalues]]]}
              if not hltl1map:
                  print 'hltl1seed mapping not found'
                  sys.exit(0)                  

              if not pargs.totable:
                  fh = pargs.ofilehandle
                  print >> fh, '# '+','.join(header)
                  csvwriter = csv.writer(fh)
              else:
                  ptable = display.create_table(header,header=True,maxwidth=60,align='l')
                                
              for hid in hltl1map.keys():
                  hdata = hltl1map[hid]
                  for pathdata in hdata:
                      hltpathname = pathdata[1]
                      seedtype = pathdata[2]
                      seedvals = pathdata[3]
                      seedstr = ' '.join([str(i) for i in seedvals])
                      display.add_row( [ '%s'%hltpathname , '%s'%seedtype, '%s'%seedstr], fh=fh, csvwriter=csvwriter, ptable=ptable )                
              if ptable:
                  display.show_table(ptable,pargs.outputstyle)         
              del ptable
          elif is_prescale:                    
              header = ['run','cmsls','prescidx']
              if pargs.hltpath is not None: 
                  header = header+['totprescval','hltpath/prescval','logic','l1bit/prescval']

              hltrunconfig = api.get_hltrunconfig(dbengine,hltconfigid=pargs.hltconfigid,hltkey=pargs.hltkey,runnum=pargs.runmin,schemaname=dbschema)  #[['runnum','hltconfigid','hltkey'],['runnum','hltconfigid','hltkey']]
              runlist = [x[0] for x in hltrunconfig]
              hltconfigids = [x[1] for x in hltrunconfig]
              presc = api.get_prescidx_change(dbengine,runlist,schemaname=dbschema)
              #{runnum:[[lslastscaler,prescidx]]
                           
              if not pargs.totable:
                  fh = pargs.ofilehandle
                  print >> fh, '# '+','.join(header)
                  csvwriter = csv.writer(fh)
              else:
                  ptable = display.create_table(header,header=True,maxwidth=60,align='l')
                  
              if pargs.hltpath:                  
                  hltl1map = api.get_hlttrgl1seedmap(dbengine,hltpath=pargs.hltpath,hltconfigids=hltconfigids,schemaname=dbschema)
                  # {hltconfigid: [[hltpathid,hltpathname,seedtype,[seedvalues]]]}
                  if not hltl1map:
                      print 'No hltpathl1seed mapping found'
                      sys.exit(0)
                      
                  selected_hltconfigids = hltl1map.keys()
                  if not selected_hltconfigids:
                      print 'No hltconfig found '
                      sys.exit(0)
                  for [runnum,hltconfigid,hltkey] in [r for r in hltrunconfig if r[1] in selected_hltconfigids]:
                      if not hltl1map.has_key(hltconfigid):
                          continue
                      mdata = hltl1map[hltconfigid]
                      for [hltpathid,hltpathname,l1seedlogic,l1c] in mdata:                          
                          if not presc.has_key(runnum):
                              display.add_row( [ '%d'%runnum, '%s'%'None', '%s'%'None', '%d'%1,'%s'%hltpathname, '%s'%l1seedlogic, '%s'%' '.join(l1c)], fh=fh, csvwriter=csvwriter, ptable=ptable )
                          else:
                              hltpathnamemap = dict( [(h[0],[h[1],h[2]]) for h in mdata] )
                              if not hltpathnamemap.has_key(hltpathid):
                                  continue
                              for [lsnum,prescidx] in presc[runnum]:
                                  hltprescval = api.get_hltprescale(dbengine,runnum,lsnum,hltconfigid,prescidx,hltpathid,schemaname=dbschema)
                                  rl = api.get_l1prescale(dbengine,runnum,lsnum,l1candidates=l1c,prescidxs=prescidx,ignorel1mask=parseresult['--ignore-mask'] ,schemaname=dbschema)
                                  #{(prescidx,l1bitname):[l1prescval,bitmask]}
                                  if not rl: continue
                                  (hltpathname,l1seedlogic) = hltpathnamemap[hltpathid]                                  
                                  l1keys = [ k for k in rl.keys() if k[0]==prescidx ]
                                  l1bitnames = [ v[1] for v in l1keys ]
                                  l1vals = [ rl[k] for k in l1keys ]                       
                                  l1prescvals = [ v[0] for v in l1vals ]
                                  l1bits = zip(l1bitnames,l1prescvals)
                                  l1inner = map(formatter.bitprescFormatter,l1bits)
                                  l1bitsStr = ' '.join(l1inner)                          
                                  hltpathStr = '/'.join([hltpathname,str(hltprescval)])
                                  totpresc = totalprescaleNEW(hltprescval,l1seedlogic,l1prescvals)
                                  display.add_row( [ '%d'%runnum, '%d'%lsnum, '%d'%prescidx, '%d'%totpresc,'%s'%hltpathStr, '%s'%l1seedlogic, '%s'%l1bitsStr], fh=fh, csvwriter=csvwriter, ptable=ptable )  
                  del hltl1map
              else:
                  if not presc:
                      print 'Not prescaled'
                      sys.exit(0)
                  for runnum in sorted(presc.keys()):
                      for plsdata in presc[runnum]:
                          lsnum = plsdata[0]
                          prescidx = plsdata[1]
                          display.add_row( [ '%d'%runnum, '%d'%lsnum, '%d'%prescidx ], fh=fh, csvwriter=csvwriter, ptable=ptable ) 
              if ptable:
                  display.show_table(ptable,pargs.outputstyle)         
                  del ptable          
          else:
              hltrunconfig = api.get_hltrunconfig(dbengine,hltconfigid=pargs.hltconfigid,hltkey=pargs.hltkey,runnum=pargs.runmin,schemaname=dbschema)
              if not hltrunconfig:
                  print 'hltconfig not found'
                  sys.exit(0)
              header = ['hltconfigid','hltkey','run']    
              if not pargs.totable:
                  fh = pargs.ofilehandle
                  print >> fh, '# '+','.join(header)
                  csvwriter = csv.writer(fh)
              else:
                  ptable = display.create_table(header,header=True,maxwidth=80,align='l')

              hltrunconfig_df = pd.DataFrame(hltrunconfig,columns=['runnum','hltconfigid','hltkey'])
              grouped = hltrunconfig_df.groupby(['hltconfigid','hltkey'])              
              for name,group in grouped:
                  hltconfigid = int(name[0])
                  hltkey = str(name[1])
                  runsStr = ','.join([str(i) for i in group['runnum'].values])
                  display.add_row( [ '%d'%hltconfigid, '%s'%hltkey , '%s'%runsStr], fh=fh, csvwriter=csvwriter, ptable=ptable )                
              if ptable:
                  display.show_table(ptable,pargs.outputstyle)         
                  del ptable                  
              del hltrunconfig_df
          if fh and fh is not sys.stdout: fh.close()
          sys.exit(0)
          
    except docopt.DocoptExit:
      raise docopt.DocoptExit('Error: incorrect input format for '+args['<command>'])            
    except schema.SchemaError as e:
      exit(e)    
    return

if __name__ == '__main__':
    brilcalc_main()
    exit(0)



class ParametersObject (object):
    '''
    collection of constants used in lumi related calculation
    '''
    def __init__ (self):
        self.NBX             = 3564  # number beam crossings
        self.numorbit        = 2**18 # 262144
        self.rotationRate    = 11245.613 # for 3.5 TeV Beam energy
        self.rotationTime    = 1 / self.rotationRate
        self.lumiSectionLen  = self.numorbit * self.rotationTime
        ##self.minBiasXsec   = 71300 # unit: microbarn

        
    def setRotationRate(self,rate):
        '''
        update the default LHC orbit frequency
        Single beam energy of 450GeV:  11245.589
        Single beam energy of 3.5TeV: 11245.613
        '''
        self.rotationRate =rate
        
    def setNumOrbit(self,numorbit):
        self.numorbit=numorbit
        
    def setNumBx(self,numbx):
        '''
        update the default number of BX
        '''
        self.NBX = numbx
        
    def calculateTimeParameters(self):
        '''Given the rotation rate, calculate lumi section length and
        rotation time.  This should be called if rotationRate is
        updated.
        '''
        self.rotationTime    = 1 / self.rotationRate
        self.lumiSectionLen  = self.numorbit * self.rotationTime
        
    def lslengthsec(self):
        '''
        Calculate lslength in sec from number of orbit and BX
        '''
        return self.lumiSectionLen 

    def avgpu(self,avglumi,ncollidingbx,minbias):
        '''
        avglumi hz/ub
        minbias ub
        '''
        if not ncollidingbx or not avglumi:
            return 0
        return avglumi/float(ncollidingbx)*float(minbias)/self.rotationRate
    
#=======================================================
#   Unit Test
#=======================================================
if __name__ == "__main__":
    p=ParametersObject()
    print p.lslengthsec()
    print p.NBX
    print p.numorbit



_timeopt_pattern = '^\d\d/\d\d/\d\d \d\d:\d\d:\d\d$|^\d{6}$|^\d{4}$'
_fillnum_pattern = '^\d{4}$'
_runnum_pattern = '^\d{6}$'
_time_pattern = '^\d\d/\d\d/\d\d \d\d:\d\d:\d\d$'
_hltpath_pattern = '^(HLT|DST|AlCa)_[\w\*\?\[\]\!]+$'
_hltconfig_pattern = '^\d+$|^[0-9a-zA-Z\_\.\/\*\?\[\]\!]+$'
_bxlist_pattern = r'(\d+)(,\s*\d+)*'
_amodetagChoices = ['PROTPHYS','IONPHYS','PAPHYS']
_beamstatusChoices = ['FLAT TOP','SQUEEZE','ADJUST','STABLE BEAMS']
_lumitypeChoices = ['HFOC','PLTZERO','PLTTK','DT','PXL','BCM1F','PLTSLINK','BCM1FUTCA','HFET','RADMON']
_applytoChoices = ['DAQ','LUMI','BKG']
_outstyle = ['tab','html','csv']
_dbtimefm = 'MM/DD/YY HH24:MI:SS.ff6'
_pydatetimefm  = '%m/%d/%y %H:%M:%S.%f'
_datetimefm = '%m/%d/%y %H:%M:%S'
_beamstatustoid = {'NO BEAM':0,'SETUP':1,'ABORT':2,'INJECTION PROBE BEAM':3,'INJECTION SETUP BEAM':4,'INJECTION PHYSICS BEAM':5,'PREPARE RAMP':6,'RAMP':7,'FLAT TOP':8,'SQUEEZE':9,'ADJUST':10,'STABLE BEAMS':11,'UNSTABLE BEAMS':12,'BEAM DUMP WARNING':13,'BEAM DUMP':14,'RAMP DOWN':15,'CYCLING':16,'RECOVERY':17,'INJECT & DUMP':18,'CIRCULATE & DUMP':19,'UNKNOWN':20}
_idtobeamstatus = dict( (v,k) for k,v in _beamstatustoid.items() )
_amodetagtoid = {'UNKNOWN':0,'PROTPHYS':1,'IONPHYS':2,'PAPHYS':3,'APPHYS':4,'TOTEMPHYS':5}
_idtoamodetag = dict( (v,k) for k,v in _amodetagtoid.items() )




#
# pretty table printer
# written by George Sakkis
# http://code.activestate.com/recipes/267662
#
import cStringIO,operator
def indent(rows,hasHeader=False,headerChar='-',delim=' | ',justify='center',
           separateRows=False,prefix='',postfix='',wrapfunc=lambda x:x):
    """
    Indents a table by column.
    - rows: A sequence of sequences of items, one sequence per row.
    - hadHeader: True if the first row consists of the column's names.
    - headerChar: Character to be used for the row separator line
    (if hasHeader==True or separateRows==True).
    - delim: The column delimiter.
    - justify: Determines how are data justified in their column.
    Valid values are 'left','right','center'.
    - separateRows: True if rows are to be separated by a line of 'headerChar's.
    - prefix: A string prepended to each printed row.
    - postfix: A string appended to each printed row.
    - wrapfunc: A function f(text) for wrapping text; each element in the table is first wrapped by this function.
    """
    #nested function
    #closure for breaking logical rows to physical, using wrapfunc
    def rowWrapper(row):
        newRows=[wrapfunc(item).split('\n') for item in row]
        #print 'newRows: ',newRows
        #print 'map result:  ',map(None,*newRows)
        #print 'rowwrapped: ',[[substr or '' for substr in item] for item in map(None,*newRows)]
        return [[substr or '' for substr in item] for item in map(None,*newRows)]
    # break each logical row into one or more physical ones
#    logicalRows = [rowWrapper(row) for row in rows]
    logicalRows = [(row,) for row in rows]#http://code.activestate.com/recipes/577615-table-indentation/
    # columns of physical rows
    columns = map(None,*reduce(operator.add,logicalRows))
    # get the maximum of each column by the string length of its items
    maxWidths = [max([len(str(item)) for item in column]) for column in columns]
    rowSeparator = headerChar * (len(prefix) + len(postfix) + sum(maxWidths) + len(delim)*(len(maxWidths)-1))
    # select the appropriate justify method
    justify = {'center':str.center,'right':str.rjust,'left':str.ljust}[justify.lower()]
    output=cStringIO.StringIO()
    if separateRows: print >> output,rowSeparator
    for physicalRows in logicalRows:
        for row in physicalRows:
            print >> output, prefix+delim.join([justify(str(item),width) for (item,width) in zip(row,maxWidths)])+postfix
        if separateRows or hasHeader: print >> output, rowSeparator; hasHeader=False
    return output.getvalue()

# word-wrap functions
# written by Mike Brown
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/148061
import math,re
def wrap_always(text, width):
    """
    A simple word-wrap function that wraps text on exactly width characters.
    It doesn't split the text in words.
    """
    return '\n'.join([ text[width*i:width*(i+1)] \
                       for i in xrange(int(math.ceil(1.*len(text)/width))) ])

def wrap_onspace(text,width):
    """
    A word-wrap function that preserves existing line breaks
    and most spaces in the text. Expects that existing line
    breaks are posix newlines (\n).
    """
    return reduce(lambda line, word, width=width: '%s%s%s' %
                  (line,
                   ' \n'[(len(line[line.rfind('\n')+1:])
                          + len(word.split('\n',1)[0]
                                ) >= width)],
                   word),
                  text.split(' ')
                  )
def wrap_onspace_strict(text, width):
    """
    Similar to wrap_onspace, but enforces the width constraint:
    words longer than width are split.
    """
    wordRegex = re.compile(r'\S{'+str(width)+r',}')
    return wrap_onspace(wordRegex.sub(lambda m: wrap_always(m.group(),width),text),width)

if __name__ == '__main__':
    labels=('First Name','Last Name','Age','Position')
    data="""John,Smith,24,Software Engineer
    Mary,Brohowski,23,Sales Manager
    Aristidis,Papageorgopoulos,28,Senior Reseacher"""
    rows=[row.strip().split(',') for row in data.splitlines()]
    print rows
    print 'without wrapping function\n'
    print 'raw input: ',[labels]+rows
    print indent([labels]+rows,hasHeader=True)
    width=10
    for wrapper in (wrap_always,wrap_onspace,wrap_onspace_strict):
        print 'Wrapping function: %s(x,width=%d)\n'%(wrapper.__name__,width)
        print indent([labels]+rows,hasHeader=True,separateRows=True,prefix='| ',postfix=' |',wrapfunc=lambda x: wrapper(x,width))

    lumidata=[\
        ('%-*s'%(8,'run'),'%-*s'%(8,'first'),'%-*s'%(8,'last'),'%-*s'%(10,'delivered'),'%-*s'%(10,'recorded'),'%-*s'%(20,'recorded\nmypathdfdafddafd')),\
        ['%d'%(132440),'%d'%(23),'%d'%(99),'%.2f'%(2.345),'%.2f'%(1.23),'%.2f'%(0.5678)],\
        ['%d'%(132442),'%d'%(1),'%d'%(20),'%.2f'%(2.345),'%.2f'%(1.23),'%.2f'%(0.5678)],\
        ['','%d'%(27),'%d'%(43),'%.2f'%(2.345),'%.2f'%(1.23),'%.2f'%(0.5678)]\
        ]
    lumiheader=[('%-*s'%(30,'Lumi Sections'),'%-*s'%(46,'Luminosity'))]
    headerwidth=46
    print indent(lumiheader,hasHeader=True,separateRows=False,prefix='| ',postfix='',wrapfunc=lambda x: wrap_always(x,headerwidth))
    lumifooter=[('%-*s'%(24,'189'),'%-*s'%(10,'17.89'),'%-*s'%(10,'16.1'),'%-*s'%(20,'3.47'))]
    width=20
    print indent(lumidata,hasHeader=True,separateRows=False,prefix='| ',postfix='',wrapfunc=lambda x: wrap_onspace_strict(x,width))
    print
    print indent(lumifooter,hasHeader=False,separateRows=True,prefix=' total: ',postfix='',delim=' | ',wrapfunc=lambda x: wrap_always(x,25))



"""

Usage:
  brilschema loadresult [options] 

Options:
  -h, --help       Show this screen
  -i INSELECT      Input run, ls selection json file or dict string 
  -o OUTPUT        Output csv file or bril database 
  -p AUTH          Path to authentication.ini 
  --name NAME      Result: hfoclumi,pixellumi
  --lumdb LUMIDB   Source lumi database or csv file

"""


import os,re
from docopt import docopt
from schema import Schema, And, Or, Use
choices_resultdata = ['hfoclumi','pixellumi']


def validate(optdict):
    result={}
    schema = Schema({ 
     '-i': str,
     '-o': Or(None,And(str,lambda s: os.path.isfile(os.path.expanduser(s)) or s.find('sqlite:')!=-1 or s.find('oracle:')!=-1) ,error='-o OUTPUT should be a csv file or database connect string'),     
     '--name': And(str,lambda s: s.lower() in choices_mapdata, error='NAME must be in '+str(choices_resultdata)),
     '--lumidb': And(str,lambda s: os.path.isfile(os.path.expanduser(s)) or s.find('sqlite:')!=-1 or s.find('oracle:')!=-1 ,error='-o LUMIDB should be a csv file or database connect string'),  
     '-p': Or(None,And(str,lambda s: os.path.isfile(os.path.join(os.path.expanduser(s),'authentication.ini'))) ,error='-p AUTH should contain authentication.ini'),   
     str:object # catch all
    })
    result=schema.validate(optdict)    
    return result

if __name__ == '__main__':
    print docopt(__doc__,options_first=True)



#!/usr/bin/env python
import sys
import os
import re

from distutils.core import setup
from setuptools import setup
with open("README.md", "rb") as f:
    long_desc = f.read().decode('utf-8')
version = re.search(
                '^__version__\s*=\s*"(.*)"',
                open('brilws/_version.py').read(),
                re.M
                ).group(1)
print version

setup(
    name = "brilws",
    author = "Zhen Xie",
    url = "https://github.com/xiezhen/brilws",
    download_url = 'https://github.com/xiezhen/brilws/tarball/'+version,
    license = "MIT",
    version = version,
    description = "bril worksuite",
    long_description = long_desc,
    packages = ['brilws', 'brilws.cli'],
    entry_points = {
        "console_scripts" : ['brilcalc = brilws.cli.brilcalc_main:brilcalc_main','brilschema = brilws.cli.brilschema_main:brilschema_main','briltag = brilws.cli.briltag_main:briltag_main']
        },
    package_data = {'data':['brilws/data/'],'dbschema':['brilws/dbschema/*.yaml']},
    include_package_data=True,
)




from brilws.cli.brilcalc_main import brilcalc_main
assert brilcalc_main
from brilws.cli.briltag_main import briltag_main
assert briltag_main
from brilws.cli.brilschema_main import brilschema_main
assert brilschema_main



from sqlalchemy.dialects.oracle import frontier
from brilws import api,params,RegexValidator
import re,time,os,sys
from datetime import datetime
import calendar
from schema import And, Or, Use
from dateutil import tz
from ConfigParser import SafeConfigParser
import yaml
import numpy as np
def parseservicemap(authfile):
    '''
    parse service config ini file
    output: {servicealias:[protocol,user,passwd,descriptor]}
    '''
    result={}
    parser = SafeConfigParser()
    parser.read(authfile)
    for s in parser.sections():
        protocol = parser.get(s,'protocol')
        user = parser.get(s,'user')
        passwd = parser.get(s,'pwd')
        descriptor = parser.get(s,'descriptor')
        result[s] = [protocol,user,passwd,descriptor]
    return result

class parser(object):
    def __init__(self,argdict):
        self._argdict = argdict
        self._dbconnect = None 
        self._authpath = None
        self._beamstatus = None
        self._egev = None
        self._datatagname = None
        self._amodetag = None
        self._fillmin = None
        self._fillmax = None
        self._runmin = None
        self._runmax = None
        self._tssecmin = None
        self._tssecmax = None
        self._runlsSeries = None
        self._iovtagSelect = None
        self._withBX = False
        self._xingMin = 0.
        self._xingTr = 0.
        self._xingId = []
        self._byls = False
        self._chunksize = None
        self._lumitype = None        
        self._hltpath = None
        self._hltkey = None
        self._hltconfigid = 0
        self._ofilename = '-'
        self._fh = None
        self._totable = False
        self._name = None
        self._comments = ''
        self._outputstyle = 'tab'
        self._applyto = ''
        self._scalefactor = 1.
        self._minbias = 0.
        self._cerntime = False
        self._tssec = False
        self._withoutcorrection = False
        self._yamlobj = None
        self._servicemap = {}
        self._parse()
        
    def _parse(self):

        self._dbconnect = self._argdict['-c']
        if self._argdict.has_key('-p'):
            self._authpath = self._argdict['-p']
            if not self._authpath:
                self._authpath = os.path.dirname(os.path.abspath(__file__))
                self._authpath = os.path.join(os.path.sep,self._authpath,'../data/readdb3.ini')
            self._servicemap = parseservicemap(self._authpath)
        if self._argdict.has_key('-b') and self._argdict['-b']:
            self._beamstatus = self._argdict['-b'].upper()            
        if self._argdict.has_key('--beamenergy'):
            self._egev = self._argdict['--beamenergy']
        if self._argdict.has_key('--minBiasXsec'):
            self._minbias = self._argdict['--minBiasXsec']    
        if self._argdict.has_key('--datatag'):
            self._datatagname = self._argdict['--datatag']
        if self._argdict.has_key('--amodetag'):
            self._amodetag = self._argdict['--amodetag']
        if self._argdict.has_key('--chunk-size'):
            self._chunksize = self._argdict['--chunk-size']
        if self._argdict.has_key('--output-style'):
            self._outputstyle = self._argdict['--output-style']
        if self._argdict.has_key('--name'):
            self._name = self._argdict['--name']
        if self._argdict.has_key('--comments'):
            self._comments = self._argdict['--comments']
        if self._argdict.has_key('--xing'):
            self._withBX = self._argdict['--xing']
        if self._argdict.has_key('--xingMin'):
            self._xingMin = self._argdict['--xingMin']
        if self._argdict.has_key('--xingTr'):
            self._xingTr = self._argdict['--xingTr']
        if self._argdict.has_key('--xingId') and self._argdict['--xingId']:
            d = self._argdict['--xingId']
            if isinstance(d,file):
                d = d.read()
            dd = np.array([int(i) for i in d.replace(' ','').split(',')])
            if not (dd>=1).all() and (dd<=3564).all():
                raise ValueError('--xingId should be in range [1,3564]')
            self._xingId = dd.tolist()
        if self._argdict.has_key('--byls'):
            self._byls = self._argdict['--byls']        
        if self._argdict.has_key('--type'):
            self._lumitype = self._argdict['--type']
        if self._argdict.has_key('--hltpath'):
            self._hltpath = self._argdict['--hltpath']
        if self._argdict.has_key('--hltconfig'):
            hltconfig = self._argdict['--hltconfig']
            if hltconfig:
                if hltconfig.isdigit():
                    self._hltconfigid = int(hltconfig)
                else:
                    self._hltkey = hltconfig
        if self._argdict.has_key('--applyto'):
            self._applyto = self._argdict['--applyto']
        if self._argdict.has_key('-y'):
            self._yamlfile = self._argdict['-y']            
        if self._argdict.has_key('-n'):
            self._scalefactor = self._argdict['-n']
        if self._argdict.has_key('--cerntime'):
            self._cerntime = self._argdict['--cerntime']
        if self._argdict.has_key('--tssec'):
            self._tssec = self._argdict['--tssec']
        if self._argdict.has_key('--without-correction'):
            self._withoutcorrection = self._argdict['--without-correction']
        if self._argdict.has_key('-f') and self._argdict['-f'] :
            self._fillmin = self._argdict['-f']
            self._fillmax = self._argdict['-f']
        if self._argdict.has_key('--normtag') and self._argdict['--normtag']:
            normtagfileorpath = self._argdict['--normtag']
            self._iovtagSelect = api.parseiovtagselectionJSON(normtagfileorpath)
        if self._argdict.has_key('-i') and self._argdict['-i']: # -i has precedance over -r
            fileorpath = self._argdict['-i']
            self._runlsSeries = api.parsecmsselectJSON(fileorpath)
        if self._argdict.has_key('-r') and self._argdict['-r'] :
            self._runmin = self._argdict['-r']
            self._runmax = self._argdict['-r']
        s_beg = None
        s_end = None
        #if self._argdict.has_key('-f') and not self._argdict['-f'] and self._argdict.has_key('-r') and not self._argdict['-r']:
        if self._argdict.has_key('-f') or self._argdict.has_key('-r') :
            if self._argdict.has_key('--begin') and self._argdict['--begin']:
                s_beg = self._argdict['--begin']
                for style,pattern in {'fill':params._fillnum_pattern,'run':params._runnum_pattern, 'time':params._time_pattern}.items():
                    if re.match(pattern,s_beg):
                        if style=='fill':
                            if self._fillmin : #-f specified
                                if self._fillmin<int(s_beg):
                                    raise ValueError('-f FILLNUM is less than --begin')                                
                                self._fillmin = max(self._fillmin,int(s_beg))
                            else:
                                self._fillmin = int(s_beg)
                        elif style=='run':
                            if self._runmin : #-r specified
                                if self._runmin<int(s_beg):
                                    raise ValueError('-r RUNNUM is less than --begin')
                                self._runmin = max(self._runmin,int(s_beg))
                            else:
                                self._runmin = int(s_beg)
                        elif style=='time':
                            dt_obj = datetime.strptime(s_beg,params._datetimefm)                            
                            self._tssecmin = calendar.timegm(dt_obj.timetuple())
            if self._argdict.has_key('--end') and self._argdict['--end']:
                s_end = self._argdict['--end']
                for style,pattern in {'fill':params._fillnum_pattern,'run':params._runnum_pattern, 'time':params._time_pattern}.items():
                      if re.match(pattern,s_end):
                        if style=='fill':
                            if self._fillmax : #-f specified
                                if self._fillmax>int(s_end):
                                    raise ValueError('-f FILLNUM is greater than --end')
                                self._fillmax = min(self._fillmax,int(s_end))
                            else:
                                self._fillmax = int(s_end)
                        elif style=='run':
                            if self._runmax : #-r specified
                                if self._runmax>int(s_end):
                                    raise ValueError('-r RUNNUM is greater than --end')
                                self._runmax = min(self._runmax,int(s_end))
                            else:
                                self._runmax = int(s_end)
                        elif style=='time':
                            dt_obj = datetime.strptime(s_end,params._datetimefm)
                            self._tssecmax = calendar.timegm(dt_obj.timetuple())
        
        if self._argdict.has_key('-o') and self._argdict['-o'] or self._outputstyle == 'csv':
            if self._argdict['-o']:
                self._outputstyle = 'csv'                
                self._ofilename = self._argdict['-o']
                self._fh = open(self._ofilename,'w')                
            else:
                self._fh = sys.stdout
        else:
            self._totable = True
        
    @property
    def dbconnect(self):
        return self._dbconnect
    @property
    def authpath(self):
        return self._authpath
    @property
    def beamstatus(self):
        return self._beamstatus
    @property
    def beamstatusid(self):
        if not self._beamstatus: return None
        return params._beamstatustoid[self._beamstatus]
    @property
    def egev(self):
        return self._egev
    @property
    def datatagname(self):
        return self._datatagname
    @property
    def amodetag(self):
        return self._amodetag
    @property
    def amodetagid(self):
        if not self._amodetag: return None
        return params._amodetagtoid[self._amodetag]
    @property
    def fillmin(self):
        return self._fillmin
    @property
    def fillmax(self):
        return self._fillmax
    @property
    def runmin(self):
        return self._runmin
    @property
    def runmax(self):
        return self._runmax
    @property
    def tssecmin(self):
        return self._tssecmin
    @property
    def tssecmax(self):
        return self._tssecmax
    @property
    def runlsSeries(self):
        return self._runlsSeries
    @property
    def iovtagSelect(self):
        return self._iovtagSelect
    @property
    def withBX(self):
        return self._withBX
    @property
    def byls(self):
        return self._byls
    @property
    def chunksize(self):
        return self._chunksize
    @property
    def ofilehandle(self):
        return self._fh
    @property
    def outputstyle(self):
        return self._outputstyle
    @property
    def totable(self):
        return self._totable        
    @property
    def name(self):
        return self._name
    @property
    def comments(self):
        return self._comments
    @property
    def lumitype(self):
        return self._lumitype
    @property
    def hltpath(self):
        return self._hltpath
    @property
    def hltconfigid(self):
        return self._hltconfigid
    @property
    def hltkey(self):
        return self._hltkey
    @property
    def applyto(self):
        return self._applyto
    @property
    def scalefactor(self):
        return self._scalefactor
    @property
    def cerntime(self):
        return self._cerntime
    @property
    def tssec(self):
        return self._tssec
    @property
    def minbias(self):
        return self._minbias
    @property
    def xingMin(self):
        return self._xingMin
    @property
    def xingTr(self):
        return self._xingTr
    @property
    def xingId(self):
        return self._xingId
    @property
    def withoutcorrection(self):
        return self._withoutcorrection

    @property
    def connecturl(self):
        if not os.path.isfile(self._dbconnect):
            if not self._servicemap.has_key(self._dbconnect):
                raise ValueError('service %s is not defined'%self._dbconnect)
            protocol = self._servicemap[self._dbconnect][0]
            if protocol not in ['oracle','frontier'] : raise ValueError('protocol %s is not supported'%protocol)       
            descriptor = self._servicemap[self._dbconnect][3]
            if protocol == 'oracle':
                user = self._servicemap[self._dbconnect][1]
                passwd = self._servicemap[self._dbconnect][2].decode('base64')
                connecturl = 'oracle+cx_oracle://%s:%s@%s'%(user,passwd,descriptor)
            else:                
                connecturl = 'oracle+frontier://%s'%(descriptor)
        else:  #only webconfig frontier takes file as -c argument
            connecturl = 'oracle+frontier:///%s/%s'%(self._dbconnect,'LumiCalc')            
        return connecturl        
        
    @property
    def yamlfile(self):
        return self._yamlfile
    
    @property
    def yamlobj(self):
        with open(self._yamlfile,'r') as f:
            self._yamlobj = yaml.safe_load(f)
        return self._yamlobj
    
argvalidators = {
    '--amodetag': Or(None,And(str,lambda s: s.upper() in params._amodetagChoices), error='--amodetag must be in '+str(params._amodetagChoices) ),
    '--beamenergy': Or(None,And(Use(int), lambda n: n>0), error='--beamenergy should be a positive number'),
    '--xingMin': Or(None,And(Use(float), lambda n: n>0), error='--xingMin should be a positive number'),
    '--xingTr': Or(None,And(Use(float), lambda n: (n>0 and n<=1)), error='--xingTr should be a number in (0,1]'),
    '--xingId': Or(None,Or( Use(open),Use(RegexValidator.RegexValidator(params._bxlist_pattern))), error='--xingId should be a comma separated list of numbers'),
    '--minBiasXsec': And(Use(float), lambda f: f>0, error='--minBiasXsec should be float > 0'),
    '-b': Or(None, And(str, lambda s: s.upper() in params._beamstatusChoices), error='-b must be in '+str(params._beamstatusChoices) ),
    '--begin': Or(None, And(str,Use(RegexValidator.RegexValidator(params._timeopt_pattern))), error='--begin wrong format'),
    '--end': Or(None, And(str,Use(RegexValidator.RegexValidator(params._timeopt_pattern))), error='--end wrong format'),
    '--output-style': And(str,Use(str.lower), lambda s: s in params._outstyle, error='--output-style choice must be in '+str(params._outstyle) ),
#    '--chunk-size':  And(Use(int), lambda n: n>0, error='--chunk-size should be integer >0'),
    '--type': Or(None, And(str, lambda s: s.upper() in params._lumitypeChoices), error='--type must be in '+str(params._lumitypeChoices) ),
    '--hltpath': Or(None, And(str, Use(RegexValidator.RegexValidator(params._hltpath_pattern))),  error='--hltpath wrong format'),
    '--hltconfig': Or(None, And(str, Use(RegexValidator.RegexValidator(params._hltconfig_pattern))),  error='--hltconfig wrong format'),
    '--applyto': Or(None, And(str, lambda s: s.upper() in params._applytoChoices), error='--applyto must be in '+str(params._applytoChoices) ),
    '--siteconfpath': Or(None, str, error='--siteconfpath should be string'),
    '-c': str,
    '-p': Or(None,os.path.exists, error='AUTHPATH should exist'),
    '-y': And(os.path.exists, error='YAMLFILE should exist'),
    '--normtag': Or(None,str),
    '-i': Or(None,str),
    '-o': Or(None,str),    
    '-f': Or(None, And(Use(RegexValidator.RegexValidator(params._fillnum_pattern)),Use(int)), error='-f wrong format'), 
    '-n': And(Use(float), lambda f: f>0, error='-n SCALEFACTOR should be float >0'),      
    '-r': Or(None, And(Use(RegexValidator.RegexValidator(params._runnum_pattern)),Use(int)), error='-r wrong format'),
    str:object # catch all
}
